"""
A python module for dealing with terminal-based inputs and outputs.
"""

__version__ = "0.0.3"
__author__ = "Vivaan Singhvi"

from .menu import *
from .input import *
from .style import *
from .loading import *
from .printing import *